/*
 *	Title:		o3find
 *	Version:	0.8
 *	Created by:	Paolo Copello
 *	Home page:	http://web.tiscali.it/fanelia
 *	
 *	Copyright (C) 2003-2005 Paolo Copello
 *	
 *	******************************************
 *	This file is part of o3find
 *	******************************************
 *	
 *	o3find is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	any later version.
 *	
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *	
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *	
 */



// oofind.cpp : Defines the entry point for the console application.
//


// suppress "truncated debug info" warnings (VC++ specific)
#ifdef _MSC_VER
#pragma warning (disable: 4786)
#pragma warning (disable: 4788)
#endif // _MSC_VER

#include <iostream>
using namespace std;

#include "../common/config.h"

//#include "../common/o3findclass.h"

#include "../common/fnutils.h"

#include "../common/fnfilefind.h"

#include "o3findconsoleclass.h"
#include "CmdLine.h"

// options strings

const char O3FIND_OPT_DIR[]				= "-d";
const char O3FIND_OPT_SEARCHSTRING[]	= "-s";

const char O3FIND_OPT_SUBDIR[]			= "-r";
const char O3FIND_OPT_IGNORECASE[]		= "-i";

const char O3FIND_OPT_IGNOREERRORS[]	= "-e";

const char O3FIND_OPT_SHOWNOTFOUND[]	= "-nf";


const char O3FIND_OPT_FILES_STDIN[]		= "-input";

const char O3FIND_OPT_FILES_LIST[]		= "-f";

const char O3FIND_OPT_BRIEF[]		= "-b";

const char O3FIND_OPT_META[]		= "-meta";
const char O3FIND_OPT_METAONLY[]	= "-metaonly";

const char O3FIND_OPT_EXECUTE[]		= "-exec";
//const char O3FIND_OPT_OPEN[]		= "-open";

const char O3FIND_OPT_HELP1[]		= "-help";
const char O3FIND_OPT_HELP2[]		= "--help";

const char O3FIND_OPT_READ_ALREADY_OPEN_FILE[] = "-ao";

/** Read a list of file names from standard input. The names must 
be separated by newlines */
void ReadFilesFromStdin(list<string>& lst)
{
	char buffer[O3FIND_MAX_BUFFER_SIZE];
	while(cin.getline(buffer, O3FIND_MAX_BUFFER_SIZE-1))
	{
		if(strlen(buffer)>0)
			lst.push_back(buffer);
	}
}


/** Show basic usage and options  */
void UsageBasic()
{
	cout << "o3find - Version 0.8\n"
		<< "Copyright (C) 2003-2005 Paolo Copello\n"
		<< "Can be distributed under the terms of the GNU GPL.\n\n"
		<< "Usage: \n"
		<< "\t o3find [options] -s <searchstring> \n\n"
		<< "  <searchstring> The text you wish to find.\n"
		<< "\n"
		<< "Options:\n"
		<< " -d <directory> [<directory>...]\n"
		<< "       Directory(es) to search (the end slash is not needed)\n"
		<< "       If -d is omitted, the current directory is searched\n"
		<< " -i             Ignore case \n"
		<< " -r             Search subdirectories \n"
		<< "\n"
		<< " -help          Display advanced options and examples\n" << endl;
}

/** Show advanced options and examples  */
void UsageAdvanced()
{
	UsageBasic();

#ifdef FANEL_O3FIND_USE_WAIT_FOR_ENTER
	cout << "Press ENTER to continue." << endl;
	Fanel::WaitForEnter();
#endif

	cout << "Advanced options:\n"
		<< " -f <pattern> [<pattern>...] Find specified file name(s)/pattern(s)\n"
		<< " -e         Do not show errors and warnings \n"
		<< " -b         Brief mode (do not show matching lines)\n"
		<< " -nf        Show files where the search text was not found \n"
		<< " -ao        Read also files that are already open by other programs \n"
		<< " -meta      Search metadata\n"
		<< " -metaonly  Search metadata only (ignore document content)\n"
		<< " -exec <command> Execute the command for each found file \n"
		<< "      The symbol {} in <command> will be replaced with the found file name\n"
		<< " -input     Read files to find from the standard input\n"
		<< "      If this option is specified, -d and -r are ignored\n" << endl;

#ifdef FANEL_O3FIND_USE_WAIT_FOR_ENTER
	cout << "Press ENTER to continue." << endl;
	Fanel::WaitForEnter();
#endif

	cout << "Examples:\n"
		<< "\t o3find -d d:\\mydocs -s dog  \n"
		<< "\t o3find -d d:\\mydocs\\ -s \"cats and dogs\" \n"
		<< "\t o3find \"c:/documents and settings/jack/\" -s dog  \n"
		<< "\n"
		<< "NOTE: special characters like accented letters are not supported on Windows 95.\n"
		<< "\t e.g. To search for \"perch\" on Windows 95, type \"perch\"\n"
		<< "By default, the program searches only in .sx? files.\n"
		<< "You can use -f *.* to search all files. For non-OOo files, matching lines are not shown." << endl;
}


/** Search and replace a substring in a string */
void srch_repl(
  std::string &s,
  const std::string &to_find,
  const std::string& repl_with
) {
  std::string result;
  std::string::size_type pos = 0;
  while(true) {
    std::string::size_type next = s.find(to_find, pos);
    result.append(s, pos, next-pos);
    if( next != std::string::npos ) {
      result.append(repl_with);
      pos = next + to_find.size();
    } else {   break;  // exit loop
    }
  } s.swap(result);
}

//void OpenFoundFiles(const list<string>& lst)
//{
//#ifdef WIN32
//    SHELLEXECUTEINFO shExecInfo;
//
//    shExecInfo.cbSize = sizeof(SHELLEXECUTEINFO);
//    shExecInfo.fMask = NULL;
//    shExecInfo.hwnd = NULL;
//    shExecInfo.lpVerb = NULL;
////    shExecInfo.lpFile = pszParseName;
//    shExecInfo.lpParameters = NULL;
//    shExecInfo.lpDirectory = NULL;
//    shExecInfo.nShow = SW_SHOW;
//    shExecInfo.hInstApp = NULL;
//
//	for(list<string>::const_iterator it = lst.begin(); it != lst.end() ; ++it)
//	{
//		shExecInfo.lpFile = it->c_str();
//		ShellExecuteEx(&shExecInfo);
//		strName.ReleaseBuffer();
//	}
//
//#endif
//	cerr << "-Open operation unsupported." << endl;
//}

/** o3find console entry point */
int main(int argc, char* argv[])
{
	CCmdLine cmd;
	cmd.SplitLine(argc, argv);

	if(cmd.HasSwitch(O3FIND_OPT_HELP1) || cmd.HasSwitch(O3FIND_OPT_HELP2))
	{
		UsageAdvanced();
		return 0;
	}

	bool bCaseSensitive = true;
	bool bSearchSubdirs	= false;
	bool bShowErrors = true;
	bool bShowNotFound = false;
	bool bBriefMode = false;
	bool bReadAlreadyOpenFile = false;

	string sSearch;
	list<string> directoryList;

	if (cmd.HasSwitch(O3FIND_OPT_IGNORECASE))
		bCaseSensitive = false;
	if (cmd.HasSwitch(O3FIND_OPT_SUBDIR))
		bSearchSubdirs = true;
	if (cmd.HasSwitch(O3FIND_OPT_IGNOREERRORS))
		bShowErrors = false;
	if (cmd.HasSwitch(O3FIND_OPT_SHOWNOTFOUND))
		bShowNotFound = true;
	if (cmd.HasSwitch(O3FIND_OPT_BRIEF))
		bBriefMode = true;
	if (cmd.HasSwitch(O3FIND_OPT_READ_ALREADY_OPEN_FILE))
		bReadAlreadyOpenFile = true;


	if (cmd.GetArgumentCount(O3FIND_OPT_SEARCHSTRING) > 0)
	{
		sSearch = cmd.GetSafeArgument(O3FIND_OPT_SEARCHSTRING, 0, "");
		if(sSearch.empty())
		{
			if(bShowErrors)
				cerr << "-Error: search text unspecified." << endl;
			return 1;
		}
	}
	else
	{
		UsageBasic();
		return 1;
	}

	// check conflicting options
	if (cmd.HasSwitch(O3FIND_OPT_FILES_LIST) && cmd.HasSwitch(O3FIND_OPT_FILES_STDIN))
	{
		if(bShowErrors)
			cerr << "-Error: " << O3FIND_OPT_FILES_LIST << " and " 
				 << O3FIND_OPT_FILES_STDIN << " cannot be used together." << endl;
		return 1;
	}
	if (cmd.HasSwitch(O3FIND_OPT_META) && cmd.HasSwitch(O3FIND_OPT_METAONLY))
	{
		if(bShowErrors)
			cerr << "-Error: " << O3FIND_OPT_META << " and " 
				 << O3FIND_OPT_METAONLY << " cannot be used together." << endl;
		return 1;
	}
//	if (cmd.HasSwitch(O3FIND_OPT_EXECUTE) && cmd.HasSwitch(O3FIND_OPT_OPEN))
//	{
//		if(bShowErrors)
//			cerr << "-Error: " << O3FIND_OPT_EXECUTE << " and " 
//				 << O3FIND_OPT_OPEN << " cannot be used together." << endl;
//		return 1;
//	}

	// search content, metadata or both. By default, content.
	Fanel::OoFinder::SEARCH_IN search_in = Fanel::OoFinder::CONTENT_ONLY;

	if (cmd.HasSwitch(O3FIND_OPT_META))
		search_in = Fanel::OoFinder::CONTENT_AND_META;
	else if(cmd.HasSwitch(O3FIND_OPT_METAONLY))
		search_in = Fanel::OoFinder::META_ONLY;


	// read file names from...

	list<string> fileList;

	if (cmd.HasSwitch(O3FIND_OPT_FILES_STDIN))
	{
		// this option reads file names from standard input

		if (bShowErrors && bSearchSubdirs)
			cerr << "-Warning: " << O3FIND_OPT_FILES_STDIN << " specified, " << O3FIND_OPT_SUBDIR << " ignored." << endl;

		if (bShowErrors && cmd.GetArgumentCount(O3FIND_OPT_DIR) > 0)
			cerr << "-Warning: " << O3FIND_OPT_FILES_STDIN << " specified, " << O3FIND_OPT_DIR << " ignored." << endl;

		ReadFilesFromStdin(fileList);
	}
	else
	{
		// find files in directories
		int nDirs = cmd.GetArgumentCount(O3FIND_OPT_DIR);
		
		if(nDirs > 0)
		{
			string sDirectory;
			for(int i = 0; i < nDirs; i++)
			{
				sDirectory = cmd.GetSafeArgument(O3FIND_OPT_DIR, i, "");
				if(!sDirectory.empty())
					directoryList.push_back(sDirectory);
			}
			if(directoryList.empty() && bShowErrors )
				cerr << "-Warning: unspecified search directory, using current directory." << endl;
		}

		list<string> filterList;
		int nFilesCount = cmd.GetArgumentCount(O3FIND_OPT_FILES_LIST);

		for(int k = 0; k < nFilesCount; k++)
		{
			string s = cmd.GetSafeArgument(O3FIND_OPT_FILES_LIST, k, "");
			if (!s.empty())
				filterList.push_back(s);
		}

		if(filterList.empty())
			filterList.push_back("*.sx?");	// Default: search SO/OOo files

		if(directoryList.empty())
			directoryList.push_back("");	// Default: search current directory

		Fanel::FileFinder ffinder;

		ffinder.Find(directoryList, filterList, fileList, bSearchSubdirs);
	}

	if(fileList.empty())
	{
		if(bShowErrors)
			cerr << "-No files found." << endl;
		return 1;
	}

	// Files found. Now search the searchstring.

	Fanel::OoFinderConsole finder(bCaseSensitive, bShowErrors, bShowNotFound, 
		bBriefMode, search_in, bReadAlreadyOpenFile);

	finder.Find(fileList, sSearch);

	fflush(stdin);
	fflush(stdout);

	// if an "open" option was specified, open all the documents found
//	if (cmd.HasSwitch(O3FIND_OPT_OPEN))
//	{
//		OpenFoundFiles(finder.GetResultList());
//	}

	// if an "exec" option was specified, execute the command for each file
	if (cmd.GetArgumentCount(O3FIND_OPT_EXECUTE) > 0)
	{
		string sCommand = cmd.GetSafeArgument(O3FIND_OPT_EXECUTE, 0, "");
		
		string::size_type nPos = sCommand.find("{}");
		if(nPos >= 0 && nPos < sCommand.length())
		{
			const list<string>& lst = finder.GetResultList();

			for(list<string>::const_iterator it = lst.begin(); it != lst.end() ; ++it)
			{
				string sActualCommand = sCommand;
				srch_repl(sActualCommand, "{}", "\"" + *it + "\"");
				system(sActualCommand.c_str());
			}
		}
	}

	return 0;
}


